"""
Information
---------------------------------------------------------------------
Name        : profiler.py
Location    : ~/
Author      : Tom Eleff
Published   : 2024-03-19
Revised on  : .

Description
---------------------------------------------------------------------
Contains utility functions for profiling functions.
"""

import time
import datetime as dt
from typing import Callable


# Decorator function(s)
def run_time(
    func: Callable
) -> Callable:
    """ Prints the run time of the {func} passed.

    Parameters
    ----------
    func : `Callable`
        Function object.
    """

    def wrapper(*args, **kwargs):
        t1 = time.time()
        result = func(*args, **kwargs)
        t2 = time.time()
        td = dt.timedelta(seconds=(t2-t1))
        print("\nINFO: Function {%s()} executed in %s hh:mm:ss" % (
                func.__name__,
                td
            )
        )
        return result
    return wrapper



"""
Information
---------------------------------------------------------------------
Name        : test_profiler.py
Location    : ~/tests
Author      : Tom Eleff
Published   : 2024-03-19
Revised on  : .

Description
---------------------------------------------------------------------
Tests methods within `pytilities.profiler`.
"""

import time
from pytensils import profiler


# Profiler function(s)
@profiler.run_time
def test_run_time_success():
    time.sleep(1)



"""
Information
---------------------------------------------------------------------
Name        : test_config.py
Location    : ~/tests
Author      : Tom Eleff
Published   : 2024-03-19
Revised on  : .

Description
---------------------------------------------------------------------
Tests methods within `pytilities.config`.
"""

import os
import pytest
from pytensils import config


# Config management function(s)
@pytest.fixture
def config_object() -> config.Handler:
    return config.Handler(
        path=os.path.join(
            os.path.abspath(
                os.path.dirname(__file__)
            ),
            'resources'
        ),
        file_name='config.json'
    )


@pytest.fixture
def dtype_object() -> config.Handler:
    return config.Handler(
        path=os.path.join(
            os.path.abspath(
                os.path.dirname(__file__)
            ),
            'resources'
        ),
        file_name='dtypes.json'
    )


@pytest.fixture
def config_invalid_object() -> config.Handler:
    return config.Handler(
        path=os.path.join(
            os.path.abspath(
                os.path.dirname(__file__)
            ),
            'resources'
        ),
        file_name='config_invalid.json'
    )


@pytest.fixture
def config_validation_error_object() -> config.Handler:
    return config.Handler(
        path=os.path.join(
            os.path.abspath(
                os.path.dirname(__file__)
            ),
            'resources'
        ),
        file_name='config_validation_error.json'
    )


def test_init_success(config_object):
    assert config_object.path == os.path.join(
        os.path.abspath(
            os.path.dirname(__file__)
        ),
        'resources'
    )
    assert config_object.file_name == 'config.json'


def test_init_oserror():
    with pytest.raises(OSError):
        _ = config.Handler(
            path=os.path.join(
                os.path.abspath(
                    os.path.dirname(__file__)
                ),
                'path-does-not-exist'
            ),
            file_name='config.json'
        )


def test_init_filenotfounderror():
    with pytest.raises(FileNotFoundError):
        _ = config.Handler(
            path=os.path.join(
                os.path.abspath(
                    os.path.dirname(__file__)
                ),
                'resources'
            ),
            file_name='file-not-found.json'
        )


def test_read_success(config_object):
    assert config_object.read() == {
        "config": {
            "str": "ABC",
            "bool": True,
            "int": 1,
            "float": 9.9,
            "list": ["A", "B", "C"],
            "dict": {
                "A": "a",
                "B": "b",
                "C": "c"
            }
        }
    }


def test_read_ioerror(config_invalid_object):
    with pytest.raises(IOError):
        config_invalid_object.read()


def test_validate_success(config_object, dtype_object):
    assert config.validate(
        config=config_object.read(),
        dtype=dtype_object.read()
    )


def test_validate_typeerror(config_validation_error_object, dtype_object):
    with pytest.raises(TypeError):
        config.validate(
            config=config_validation_error_object.read(),
            dtype=dtype_object.read()
        )


def test_write_success(tmp_path):

    # Create an empty config-file
    with open(os.path.join(tmp_path, 'config_temp.json'), 'w') as file:
        file.write('')

    # Output content to the config-file
    temp = config.Handler(
        path=tmp_path,
        file_name='config_temp.json'
    )
    temp.write(
        config={
            "config": {
                "str": "ABC",
                "bool": True,
                "int": 1,
                "float": 9.9,
                "list": ["A", "B", "C"],
                "dict": {
                    "A": "a",
                    "B": "b",
                    "C": "c"
                }
            }
        }
    )
    assert temp.read() == {
        "config": {
            "str": "ABC",
            "bool": True,
            "int": 1,
            "float": 9.9,
            "list": ["A", "B", "C"],
            "dict": {
                "A": "a",
                "B": "b",
                "C": "c"
            }
        }
    }



"""
Information
---------------------------------------------------------------------
Name        : test_utils.py
Location    : ~/tests
Author      : Tom Eleff
Published   : 2024-03-19
Revised on  : .

Description
---------------------------------------------------------------------
Tests methods within `pytilities.utils`.
"""

import os
import pytest
from pytensils import utils


# Directory management function(s)
def test_generate_output_directory_success(tmp_path):
    utils.generate_output_directory(
        path=tmp_path,
        root='root',
        sub_folders=['A', 'B', 'C']
    )
    assert os.path.isdir(os.path.join(tmp_path, 'root'))
    assert os.path.isdir(os.path.join(tmp_path, 'root', 'A'))
    assert os.path.isdir(os.path.join(tmp_path, 'root', 'B'))
    assert os.path.isdir(os.path.join(tmp_path, 'root', 'C'))


def test_generate_output_directory_oserror():
    with pytest.raises(OSError):
        utils.generate_output_directory(
            path=os.path.abspath(
                os.path.dirname(__file__)
            ),
            root='resources'
        )


# Data-type parsing function(s)
def test_as_type_success_str():
    assert utils.as_type(value='ABC') == 'ABC'


def test_as_type_success_int():
    assert utils.as_type(value='1', return_dtype='int') == 1


def test_as_type_success_float():
    assert utils.as_type(value='9.9', return_dtype='float') == 9.9


def test_as_type_success_bool():
    assert utils.as_type(value='True', return_dtype='bool')


def test_as_type_success_list():
    assert utils.as_type(
        value='["A", "B", "C"]',
        return_dtype='list'
    ) == ['A', 'B', 'C']


def test_as_type_success_dict():
    assert utils.as_type(
        value='{"A": "a", "B": "b", "C": "c"}',
        return_dtype='dict'
    ) == {
        'A': "a",
        'B': 'b',
        'C': 'c'
    }


def test_as_type_failure_typeerror():
    with pytest.raises(TypeError):
        utils.as_type(
            value='FALSE',
            return_dtype='bool'
        )


def test_as_type_failure_jsondecoder():
    with pytest.raises(TypeError):
        utils.as_type(
            value='{"A": "a", "B"="b", "C": "c"}',
            return_dtype='dict'
        )


def test_as_type_failure_nameerror():
    with pytest.raises(NameError):
        utils.as_type(
            value='ABC',
            return_dtype='string'
        )


def test_as_type_failure_typeerror_general():
    with pytest.raises(TypeError):
        utils.as_type(
            value='ABC',
            return_dtype='int'
        )



"""
Information
---------------------------------------------------------------------
Name        : utils.py
Location    : ~/
Author      : Tom Eleff
Published   : 2024-03-19
Revised on  : .

Description
---------------------------------------------------------------------
Contains utility functions for managing directories and data-types.
"""

import os
import ast
import json
from typing import Union


# Directory management function(s)
def generate_output_directory(
    path: str,
    root: str,
    sub_folders: Union[list, None] = None
):
    """ Generates an output directory with subfolders.

    Parameters
    ----------
    path : `str`
        Directory path to the folder that will contain the output folder
            directories.
    root : `str`
        The name of the root folder of the directory to create within `path`.
    sub_folders: `list`
        A list of folders to create within ~/`path`/`root`
    """

    if root not in os.listdir(path):

        # Create root output directory
        os.mkdir(os.path.join(path, root))

        # Create output sub-directories
        if sub_folders:
            for folder in sub_folders:
                os.mkdir(os.path.join(path, root, folder))
    else:
        raise OSError(
            '{~/%s} already exists.' % (root)
        )


# Data-type parsing function(s)
def as_type(
    value: str,
    return_dtype: str = 'str'
) -> Union[str, int, float, bool, list, dict]:
    """ Returns `value` as `return_dtype`.

    Parameters
    ----------
    value : `str`
        String of the value to convert to `return_dtype`.
    return_dtype : `str`
        Name of the datatype (`str`, `int`, `float`, `bool`, `list`, `dict`) of
            the returned value. If the returned value cannot be converted
            to `return_dtype` then a `TypeError` is raised. If the name of the
            `return_dtype` is invalid, then a `NameError` is returned.
    """

    try:
        if return_dtype.strip().upper() == 'STR':
            return str(value)

        elif return_dtype.strip().upper() == 'INT':
            return int(value)

        elif return_dtype.strip().upper() == 'FLOAT':
            return float(value)

        elif return_dtype.strip().upper() == 'BOOL':

            try:
                return ast.literal_eval(value)
            except (SyntaxError, ValueError):
                raise TypeError(
                    ' '.join([
                        "{%s} value" % (
                            value
                        ),
                        "cannot be converted to {%s}." % (
                            return_dtype
                        )
                    ])
                )

        elif (
            (return_dtype.strip().upper() == 'LIST')
            or (return_dtype.strip().upper() == 'DICT')
        ):
            try:
                return json.loads(value)
            except json.decoder.JSONDecodeError:
                raise TypeError(
                    ' '.join([
                        "{%s} value" % (
                            value
                        ),
                        "cannot be converted to {%s}." % (
                            return_dtype
                        )
                    ])
                )
        else:
            raise NameError(
                'Invalid return datatype {%s}.' % (
                    return_dtype
                )
            )
    except ValueError:
        raise TypeError(
            ' '.join([
                "{%s} value" % (
                    value
                ),
                "cannot be converted to {%s}." % (
                    return_dtype
                )
            ])
        )



"""
Information
---------------------------------------------------------------------
Name        : config.py
Location    : ~/
Author      : Tom Eleff
Published   : 2024-03-19
Revised on  : .

Description
---------------------------------------------------------------------
Contains the `class` methods for managing configuration.
"""

import os
import json


class Handler():

    def __init__(
        self,
        path: str,
        file_name: str
    ):
        """ Initializes an instance of the configuration-handler class.

        Parameters
        ----------
        path : `str`
            Directory path to the folder that contains the `file_name` of the
                '.json' config-file.
        file_name : `str`
            File name of the '.json' config-file.
        """

        # Validate the file-path
        if os.path.isdir(path):
            self.path = path
        else:
            raise OSError(
                '{~/%s} does not exist.' % (path)
            )

        # Validate the file-name
        if os.path.isfile(
            os.path.join(path, file_name)
        ):
            self.file_name = file_name
        else:
            raise FileNotFoundError(
                '{%s} does not exist within {~/%s}.' % (
                    file_name,
                    path
                )
            )

    def read(self) -> dict:
        """ Reads a '.json' config-file and returns the contents as a `dict`
        object.
        """
        with open(
            os.path.join(
                self.path,
                self.file_name
            ),
            mode='r'
        ) as file:
            try:
                return json.load(file)
            except json.decoder.JSONDecodeError:
                raise IOError(
                    "{~/%s} is not a valid '.json' config-file." % (
                        os.path.basename(
                            self.file_name
                        )
                    )
                )

    def write(
        self,
        config: dict
    ):
        """ Writes a '.json' config-file with the contents of `config`.

        Parameters
        ----------
        config : `dict`
            Dictionary object that contains the parameters essential to the
                application.
        """
        with open(
            os.path.join(
                self.path,
                self.file_name
            ),
            mode='w+'
        ) as file:
            json.dump(
                config,
                file,
                indent=4
            )


def validate(
    config: dict,
    dtype: dict
):
    """ Validates `config` against the dtypes in `dtype`.

    Parameters
    ----------
    config : `dict`
        Dictionary object that contains the parameters essential to the
            application.
    dtype : `dict`
        Dictionary object that contains the expected `config` value dtypes.
    """

    config_errors = {}
    error = False

    for section in config.keys():
        if section not in dtype.keys():
            dtype[section] = {}
            config_errors[section] = {}
        else:
            config_errors[section] = {}

        for key, value in config[section].items():
            if key not in dtype[section].keys():
                config_errors[section][key] = (
                    'No dtype found in {dtypes}.'
                )
                error = True
            else:
                if type(value).__name__ != dtype[section][key]:
                    config_errors[section][key] = (
                        'Invalid dtype. Expected <%s>.' % (
                            dtype[section][key]
                        )
                    )
                    error = True
                else:
                    pass

    len0 = max(
        [len(section) for section in config_errors.keys()]
    )

    if error:
        for section in config_errors.keys():
            if len(config_errors[section].keys()) > 0:
                print('\n')
                print(
                    '*** ERROR: The following errors occurred when' +
                    ' validating the {%s} parameters.\n' %
                    (section)
                )

                len1 = max(
                    [len(key) for key in config_errors[section].keys()]
                )
                len2 = max(
                    [
                        len(value) for value in (
                            config_errors[section].values()
                        )
                    ]
                )

                print(
                    "{:<8} {:<{len0}} {:<{len1}} {:<{len2}}".format(
                        '',
                        'Section',
                        'Key',
                        'Error',
                        len0=len0+4,
                        len1=len1+4,
                        len2=len2+4
                    )
                )
                print(
                    "{:<8} {:<{len0}} {:<{len1}} {:<{len2}}".format(
                        '',
                        '-------',
                        '---',
                        '-----',
                        len0=len0+4,
                        len1=len1+4,
                        len2=len2+4
                    )
                )

                for key, value in config_errors[section].items():
                    print(
                        ("{:<8} {:<{len0}} {:<{len1}} {:<{len2}}").format(
                            '',
                            section,
                            key,
                            value,
                            len0=len0+4,
                            len1=len1+4,
                            len2=len2+4
                        )
                    )
                print(
                    '\n'
                )
            else:
                pass
        raise TypeError(
            'Validation failed.'
        )
    else:
        return True
